from typing import Optional

from langchain_openai import ChatOpenAI
from pydantic import SecretStr

from openkosmos_ai.common.model import AIBaseConfig
from openkosmos_ai.nodes.base_node import BaseStateNode


class OpenAIClientConfig(AIBaseConfig):
    base_url: str
    api_key: str
    model: Optional[str] = None
    temperature: Optional[float] = 0.7
    top_p: Optional[float] = 0.8
    presence_penalty: Optional[float] = 1.03


class OpenAIClientNode(BaseStateNode[OpenAIClientConfig]):
    def __init__(self, config: OpenAIClientConfig, tools=None):
        super().__init__(config)
        self.llm = ChatOpenAI(
            model=config.model,
            base_url=config.base_url,
            api_key=SecretStr(config.api_key),
            temperature=config.temperature,
            top_p=config.top_p,
            presence_penalty=config.presence_penalty
        )
        if tools is not None:
            self.llm = self.llm.bind_tools(tools)

    def model(self):
        return self.llm

    def invoke(self, messages, output_schema=None):
        if output_schema is None:
            return self.llm.invoke(messages)
        else:
            return self.llm.with_structured_output(output_schema).invoke(messages)

    def chat(self, messages):
        response = self.llm.invoke(messages)
        return response.content_blocks



from openkosmos_core.config import Config

from openkosmos_ai.nodes.agent import AgentConfig
from openkosmos_ai.nodes.api import ApiConfig
from openkosmos_ai.nodes.database_client import DatabaseClientConfig
from openkosmos_ai.nodes.dify_client import DifyClientConfig
from openkosmos_ai.nodes.git_repo import GitRepoConfig
from openkosmos_ai.nodes.openai_client import OpenAIClientConfig
from openkosmos_ai.nodes.rag_flow_client import RAGFlowClientConfig
from openkosmos_ai.nodes.skill import SkillConfig
from openkosmos_ai.nodes.vector_store import VectorStoreConfig
from openkosmos_ai.nodes.xinference_client import XinferenceClientConfig


class AIConfig(Config):

    def get_api_config(self, name: str | list[str] = "api") -> ApiConfig:
        return self.get(name, ApiConfig)

    def get_vectorstore_config(self, name: str | list[str] = "vectorstore") -> VectorStoreConfig:
        return self.get(name, VectorStoreConfig)

    def get_openai_config(self, name: str | list[str] = "openai") -> OpenAIClientConfig:
        return self.get(name, OpenAIClientConfig)

    def get_database_config(self, name: str | list[str] = "database") -> DatabaseClientConfig:
        return self.get(name, DatabaseClientConfig)

    def get_gitrepo_config(self, name: str | list[str] = "gitrepo") -> GitRepoConfig:
        return self.get(name, GitRepoConfig)

    def get_dify_config(self, name: str | list[str] = "dify") -> DifyClientConfig:
        return self.get(name, DifyClientConfig)

    def get_ragflow_config(self, name: str | list[str] = "ragflow") -> RAGFlowClientConfig:
        return self.get(name, RAGFlowClientConfig)

    def get_xinference_config(self, name: str | list[str] = "xinference") -> XinferenceClientConfig:
        return self.get(name, XinferenceClientConfig)

    def get_agent_config(self, name: str | list[str] = "agent") -> AgentConfig:
        return self.get(name, AgentConfig)

    def get_skill_config(self, name: str | list[str] = "skill") -> SkillConfig:
        return self.get(name, SkillConfig)



from typing import Generic, Type

from openkosmos_ai.common.model import AI_CONFIG_T


class BaseStateNode(Generic[AI_CONFIG_T]):
    _config = None

    def __init__(self, config=None):
        self._config = config

    def config(self) -> Type[AI_CONFIG_T]:
        return self._config



import os
from datetime import datetime
from typing import Literal, Optional

from langchain_core.messages import AIMessage, HumanMessage, SystemMessage
from langchain_core.prompts import ChatPromptTemplate, load_prompt, PromptTemplate
from langchain_core.prompts.string import PromptTemplateFormat
from langfuse import get_client
from langfuse.langchain import CallbackHandler


def create_tool_message(messages, content="done", tool_call_index=0, tool_call_message_index=-1):
    tool_calls = messages[tool_call_message_index].tool_calls
    return {"role": "tool", "content": content, "tool_call_id": tool_calls[tool_call_index]['id']}


def create_ai_message(content: str | list[str | dict] | None = None):
    return AIMessage(content=content)


def create_user_message(content: str | list[str | dict] | None = None):
    return HumanMessage(content=content)


def create_system_message(content: str | list[str | dict] | None = None):
    return SystemMessage(content=content)


def dump_flow_result(result, user_content_block=False):
    if isinstance(result, dict):
        for m in result["messages"]:
            if user_content_block:
                for content_blocks in m.content_blocks:
                    print("---------------------------------------------------")
                    print(content_blocks)
            else:
                m.pretty_print()
    else:
        for e in result:
            if isinstance(e, tuple):
                e[0].pretty_print()
            else:
                messages = e.get("messages")
                if messages is not None:
                    messages[-1].pretty_print()


def create_run_config(run_name: str, thread_id_pattern="%Y%m%d-%H%M%S", recursion_limit=13, user_id="dev"):
    thread_id = datetime.now().strftime(thread_id_pattern)

    return {
        "configurable": {"thread_id": thread_id},
        "callbacks": [CallbackHandler()],
        "recursion_limit": recursion_limit,
        "run_name": run_name,
        "metadata": {
            "langfuse_session_id": thread_id,
            "langfuse_user_id": user_id
        }
    }


def get_prompt(template_name: str, role: Literal["system", "ai", "human"] = "system",
               source: Literal["local", "langfuse"] = "langfuse",
               local_dir: str = None, local_ext=".yml",
               local_format: Literal["mustache", "f-string"] = "mustache",
               langfuse_type: Literal["chat", "text"] = "text", langfuse_label: Optional[str] = None,
               langfuse_cache: Optional[int] = 86400,
               langfuse_format: PromptTemplateFormat = "mustache",
               langfuse_version: Optional[int] = None,
               save_to_local=False) -> ChatPromptTemplate:
    match source:
        case "langfuse":
            langfuse_prompt = get_client().get_prompt(template_name, type=langfuse_type,
                                                      label=langfuse_label,
                                                      cache_ttl_seconds=langfuse_cache,
                                                      version=langfuse_version)
            chat_prompt = ChatPromptTemplate.from_messages(
                messages=[(role, langfuse_prompt.prompt)],
                template_format=langfuse_format
            )
            chat_prompt.metadata = {"langfuse_prompt": langfuse_prompt}

            if save_to_local:
                template_base_dir = os.getenv("KOSMOS_TEMPLATE_DIR",
                                              default="./templates") if local_dir is None else local_dir
                os.makedirs(f"{template_base_dir}/{template_name[0:template_name.rindex("/")]}",
                            exist_ok=True)
                with open(f"{template_base_dir}/{template_name}{local_ext}", "w",
                          encoding="utf-8") as output_file:
                    output_file.write(f"_type: prompt\ntemplate_format: {local_format}\ntemplate: |-\n  ")
                    output_file.write(langfuse_prompt.prompt.replace("\n", "\n  "))

            return chat_prompt
        case _:
            template_base_dir = os.getenv("KOSMOS_TEMPLATE_DIR",
                                          default="./templates") if local_dir is None else local_dir
            prompt: PromptTemplate = load_prompt(f"{template_base_dir}/{template_name}{local_ext}",
                                                 encoding="utf8")
            chat_prompt = ChatPromptTemplate.from_messages(
                messages=[(role, prompt.template)],
                template_format=prompt.template_format)
            return chat_prompt



from openkosmos_ai.common.model import AIBaseConfig
from openkosmos_ai.nodes.base_node import BaseStateNode


class PlaceholderConfig(AIBaseConfig):
    name: str


class PlaceholderNode(BaseStateNode[PlaceholderConfig]):

    def __init__(self, config: str | PlaceholderConfig = None):
        if isinstance(config, str):
            super().__init__(PlaceholderConfig(name=config))
        else:
            super().__init__(config)

    def __call__(self, *args, **kwargs):
        print("[PLACEHOLDER]", self.config().name)
        pass



from sqlalchemy import create_engine

from openkosmos_ai.common.model import AIBaseConfig
from openkosmos_ai.nodes.base_node import BaseStateNode


class DatabaseClientConfig(AIBaseConfig):
    url: str


class DatabaseClientNode(BaseStateNode[DatabaseClientConfig]):
    def __init__(self, config: DatabaseClientConfig):
        super().__init__(config)
        self.database_engine = create_engine(config.url, pool_size=10)

    def engine(self):
        return self.database_engine
